import argparse
from typing import cast
import psycopg2

# pylint: disable-next=import-error,no-name-in-module
from pyspark.dbutils import DBUtils  # type: ignore
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import to_json, col


program = "dump_discord_metadata"


def get_postgres_password(spark: SparkSession) -> str:
    dbutils = DBUtils(spark)

    token = dbutils.secrets.get(scope="postgres", key="password")

    return token


# NOTE: Refactor this funciton when the dataframe input 'df' gets to big
#       to hold in memory. For now the amount of rows in small.
def write_pg_table_with_upsert(
    df: DataFrame, pg_connection_params: dict, pg_table: str, primary_key: list[str]
):
    """
    Write a Spark DataFrame to PostgreSQL using Psycopg2 and .collect() with upsert logic.

    :param df: Spark DataFrame
    :param pg_connection_params: Dictionary with PostgreSQL connection parameters
    :param pg_table: Target table name
    """
    # Collect rows from Spark DataFrame as a list of Row objects
    rows = df.collect()

    # Get column names from DataFrame schema
    columns = [field.name for field in df.schema.fields]
    placeholders = ", ".join(["%s"] * len(columns))
    update_columns = ", ".join(
        [f"{col} = EXCLUDED.{col}" for col in columns if col not in primary_key]
    )
    insert_query = f"""
        INSERT INTO {pg_table} ({', '.join(columns)})
        VALUES ({placeholders})
        ON CONFLICT ({', '.join(primary_key)})
        DO UPDATE SET {update_columns};
    """

    conn = psycopg2.connect(**pg_connection_params)

    try:
        with conn:
            with conn.cursor() as cursor:
                for row in rows:
                    cursor.execute(insert_query, row)
    finally:
        conn.close()


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    # Postgres
    parser.add_argument(
        "--pg-host",
        dest="pg_host",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--pg-port",
        dest="pg_port",
        required=True,
        type=int,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--pg-db",
        dest="pg_db",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--pg-user",
        dest="pg_user",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--guilds-pg-table",
        dest="guilds_pg_table",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--channels-pg-table",
        dest="channels_pg_table",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--messages-pg-table",
        dest="messages_pg_table",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--reactions-pg-table",
        dest="reactions_pg_table",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--members-pg-table",
        dest="members_pg_table",
        required=True,
        help="Postgres parameter",
    )
    parser.add_argument(
        "--users-pg-table",
        dest="users_pg_table",
        required=True,
        help="Postgres parameter",
    )

    # Tables
    parser.add_argument(
        "--guilds-table",
        dest="guilds_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--channels-table",
        dest="channels_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--messages-table",
        dest="messages_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--reactions-table",
        dest="reactions_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--members-table",
        dest="members_table",
        required=True,
        help="The unity catalog table to use",
    )
    parser.add_argument(
        "--users-table",
        dest="users_table",
        required=True,
        help="The unity catalog table to use",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    pg_password = get_postgres_password(spark)

    pg_params = {
        "host": args.pg_host,
        "port": args.pg_port,
        "dbname": args.pg_db,
        "user": args.pg_user,
        "password": pg_password,
    }

    # Guilds
    guilds = (
        spark.table(args.guilds_table)
        .withColumn("id", col("source_keys.id"))
        .withColumn("features", to_json("features"))
        .drop("ingest_version", "source_keys", "ingested_on")
    )

    write_pg_table_with_upsert(
        guilds, pg_params, args.guilds_pg_table, primary_key=["id"]
    )

    # Channels
    channels = (
        spark.table(args.channels_table)
        .withColumn("id", col("source_keys.id"))
        .withColumn("parent_id", col("source_keys.parent_id"))
        .withColumn("thread_owner_id", col("source_keys.thread_owner_id"))
        .withColumn("guild_id", col("source_keys.guild_id"))
        .withColumn("permission_overwrites", to_json("permission_overwrites"))
        .withColumn("thread_metadata", to_json("thread_metadata"))
        .withColumn("available_tags", to_json("available_tags"))
        .withColumn("applied_tags", to_json("applied_tags"))
        .withColumn("default_reaction_emoji", to_json("default_reaction_emoji"))
        .drop("ingest_version", "source_keys", "ingested_on")
    )

    write_pg_table_with_upsert(
        channels, pg_params, args.channels_pg_table, primary_key=["id"]
    )

    # Messages
    messages = (
        spark.table(args.messages_table)
        .withColumn("id", col("source_keys.id"))
        .withColumn("user_id", col("source_keys.user_id"))
        .withColumn("channel_id", col("source_keys.channel_id"))
        .withColumn("message_reference", to_json(col("source_keys.message_reference")))
        .withColumn("embeds", to_json("embeds"))
        .withColumn("thread", to_json("thread"))
        .withColumn("sticker_items", to_json("sticker_items"))
        .withColumn("poll", to_json("poll"))
        .withColumn("call", to_json("call"))
        .drop("ingest_version", "source_keys", "ingested_on", "text")
    )

    write_pg_table_with_upsert(
        messages, pg_params, args.messages_pg_table, primary_key=["id"]
    )

    # Reactions
    reactions = (
        spark.table(args.reactions_table)
        .withColumn("message_id", col("source_keys.message_id"))
        .withColumn("user_id", col("source_keys.user_id"))
        .withColumn("emoji", to_json("emoji"))
        .drop(
            "ingest_version",
            "source_keys",
            "ingested_on",
        )
    )

    write_pg_table_with_upsert(
        reactions,
        pg_params,
        args.reactions_pg_table,
        primary_key=["message_id", "user_id"],
    )

    # Members
    members = (
        spark.table(args.members_table)
        .withColumn("guild_id", col("source_keys.guild_id"))
        .withColumn("user_id", col("source_keys.user_id"))
        .withColumn("roles", to_json("roles"))
        .withColumn("avatar_decoration_data", to_json("avatar_decoration_data"))
        .drop(
            "ingest_version",
            "source_keys",
            "ingested_on",
        )
    )

    write_pg_table_with_upsert(
        members,
        pg_params,
        args.members_pg_table,
        primary_key=["guild_id", "user_id"],
    )

    # Users
    users = (
        spark.table(args.users_table)
        .withColumn("id", col("source_keys.id"))
        .drop(
            "ingest_version",
            "source_keys",
            "ingested_on",
        )
    )

    write_pg_table_with_upsert(
        users, pg_params, args.users_pg_table, primary_key=["id"]
    )


if __name__ == "__main__":
    main()
